#include "Shared_string.h"

#include <algorithm>
#include <memory>

Shared_string::Shared_string()
        : ref_count_(nullptr)
        , data_(nullptr)
{ }

Shared_string::Shared_string(Owned_string&& owned)
        : view_( owned.begin() ? string_view(owned) : string_view() )
{
    if (owned.begin()) {
        ref_count_ = new size_t(1);
        data_      = owned.release();
    } else {
        ref_count_ = nullptr;
        data_      = nullptr;
    }
}

Shared_string::Shared_string(string_view sv)
        : Shared_string(Owned_string(sv))
{ }

Shared_string::Shared_string(Shared_string const& that)
        : ref_count_(that.ref_count_)
        , data_(that.data_)
        , view_(that.view_)
{
    increment_();
}

Shared_string::Shared_string(Shared_string&& that) noexcept
        : ref_count_(that.ref_count_)
        , data_(that.data_)
        , view_(that.view_)
{
    that.pod_assign_({});
}

Shared_string::~Shared_string()
{
    decrement_();
}

Shared_string& Shared_string::operator=(Shared_string const& that)
{
    decrement_();
    pod_assign_(that);
    increment_();
    return *this;
}

Shared_string& Shared_string::operator=(Shared_string&& that) noexcept
{
    decrement_();
    pod_assign_(that);
    that.pod_assign_({});
    return *this;
}

Shared_string& Shared_string::operator=(Owned_string&& that)
{
    *this = Shared_string(std::move(that));
    return *this;
}

Shared_string& Shared_string::operator=(string_view sv)
{
    *this = Shared_string(sv);
    return *this;
}

char Shared_string::operator[](size_t i) const
{
    return view_[i];
}

Shared_string Shared_string::substring(size_t start) const
{
    Shared_string result(*this);
    result.view_ = result.view_.substring(start);
    return result;
}

Shared_string Shared_string::substring(size_t start, size_t size) const
{
    Shared_string result(*this);
    result.view_ = result.view_.substring(start, size);
    return result;
}

size_t Shared_string::size() const
{
    return view_.size();
}

bool Shared_string::empty() const
{
    return view_.empty();
}

char const* Shared_string::begin() const
{
    return view_.begin();
}

char const* Shared_string::end() const
{
    return view_.end();
}

char const* Shared_string::c_str() const
{
    return begin();
}

Shared_string::operator string_view() const
{
    return view_;
}

size_t Shared_string::ref_count() const
{
    return ref_count_ ? *ref_count_ : 0;
}

void Shared_string::decrement_()
{
    if (ref_count_ && --*ref_count_ == 0) {
        delete ref_count_;
        delete [] data_;
    }
}

void Shared_string::increment_()
{
    if (ref_count_) ++*ref_count_;
}

void Shared_string::pod_assign_(Shared_string const& that)
{
    ref_count_ = that.ref_count_;
    data_      = that.data_;
    view_      = that.view_;
}

Shared_string::operator Owned_string() &&
{
    if (ref_count() == 1 && data_ && data_ == begin()) {
        Owned_string result(Owned_string::from_owned_parts(data_, size()));
        pod_assign_({});
        return result;
    }

    return Owned_string(begin(), end());
}


